/*
 * Copyright © 2016 - 2017 Copyright (c) Sprint, Inc. and others.  All rights reserved.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v1.0 which accompanies this distribution,
 * and is available at http://www.eclipse.org/legal/epl-v10.html
 */
package org.onosproject.fpcagent.util.eventStream;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ObjectNode;
import com.google.common.collect.Maps;
import org.apache.commons.lang.exception.ExceptionUtils;
import org.apache.http.entity.ContentType;
import org.apache.http.impl.nio.client.CloseableHttpAsyncClient;
import org.apache.http.nio.client.methods.HttpAsyncMethods;
import org.apache.http.nio.protocol.HttpAsyncRequestProducer;
import org.onosproject.restconf.utils.RestconfUtils;
import org.onosproject.yang.gen.v1.ietfdmmfpcagent.rev20160803.ietfdmmfpcagent.DefaultYangAutoPrefixNotify;
import org.onosproject.yang.model.DefaultModelObjectData;
import org.onosproject.yang.model.ResourceData;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Map;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedBlockingQueue;

import static org.onosproject.fpcagent.util.FpcUtil.modelConverter;
import static org.onosproject.fpcagent.util.FpcUtil.module;


/**
 * A servlet context listener that sends response (ConfigureOutput) events to clientUriToHttpClient.
 */
public class SseManager implements AutoCloseable {
    private static final Logger log = LoggerFactory.getLogger(SseManager.class);

    //key = client-id, value = configure-output
    public BlockingQueue<Map.Entry<String, DefaultYangAutoPrefixNotify>> configureQueue;
    public BlockingQueue<Map.Entry<String, String>> registerClientQueue;
    public BlockingQueue<Map.Entry<String, String>> notificationQueue;

    public Map<String, CloseableHttpAsyncClient> clientUriToHttpClient;
    public Map<String, CloseableHttpAsyncClient> clientIdToHttpClient;
    private static SseManager _instance;

    private boolean run;

    private SseManager() {
        configureQueue = new LinkedBlockingQueue<>();
        registerClientQueue = new LinkedBlockingQueue<>();
        notificationQueue = new LinkedBlockingQueue<>();
        clientUriToHttpClient = Maps.newConcurrentMap();
        clientIdToHttpClient = Maps.newConcurrentMap();
    }

    public static SseManager createInstance() {
        if (_instance == null) {
            _instance = new SseManager();
        }
        return _instance;
    }

    public static SseManager getInstance() {
        return _instance;
    }

    public void open() {
        run = true;
        Executors.newSingleThreadExecutor().submit(() -> {
            String clientUri = null;
            while (run) {
                try {
                    Map.Entry<String, DefaultYangAutoPrefixNotify> entry = configureQueue.take();
                    clientUri = entry.getKey();
                    DefaultYangAutoPrefixNotify notify = entry.getValue();

                    ResourceData dataNode = modelConverter.createDataNode(
                            DefaultModelObjectData.builder()
                                    .addModelObject(notify)
                                    .build()
                    );
                    ObjectNode jsonNodes = RestconfUtils.convertDataNodeToJson(module, dataNode.dataNodes().get(0));
                    ObjectMapper mapper = new ObjectMapper();
                    String value = mapper.writeValueAsString(jsonNodes);

                    if (clientUriToHttpClient.containsKey(clientUri)) {
                        CloseableHttpAsyncClient client = clientUriToHttpClient.get(clientUri);
                        HttpAsyncRequestProducer get = HttpAsyncMethods
                                .createPost(clientUri, value.getBytes(), ContentType.TEXT_PLAIN);

                        client.execute(get, null, null);
                    }
                } catch (Exception e) {
                    log.error(ExceptionUtils.getFullStackTrace(e));
                    clientUriToHttpClient.remove(clientUri);
                    break;
                }
            }
        });

        Executors.newSingleThreadExecutor().submit(() -> {
            String clientUri = null;
            while (run) {
                try {
                    Map.Entry<String, String> entry = registerClientQueue.take();
                    clientUri = entry.getKey();

                    if (clientUriToHttpClient.containsKey(clientUri)) {
                        CloseableHttpAsyncClient client = clientUriToHttpClient.get(clientUri);
                        HttpAsyncRequestProducer get = HttpAsyncMethods
                                .createPost(clientUri, entry.getValue().getBytes(), ContentType.TEXT_PLAIN);
                        client.execute(get, null, null);
                    }
                } catch (Exception e) {
                    log.error(ExceptionUtils.getFullStackTrace(e));
                    clientUriToHttpClient.remove(clientUri);
                    break;
                }
            }

        });

        Executors.newSingleThreadExecutor().submit(() -> {
            String clientId = null;
            while (run) {
                try {
                    Map.Entry<String, String> entry = notificationQueue.take();
                    clientId = entry.getKey();

                    if (clientIdToHttpClient.containsKey(clientId)) {
                        CloseableHttpAsyncClient client = clientIdToHttpClient.get(clientId);
                        HttpAsyncRequestProducer get = HttpAsyncMethods
                                .createPost(clientId, entry.getValue().getBytes(), ContentType.TEXT_PLAIN);
                        client.execute(get, null, null);
                    }
                } catch (Exception e) {
                    log.error(ExceptionUtils.getFullStackTrace(e));
                    clientIdToHttpClient.remove(clientId);
                    break;
                }
            }

        });
    }

    @Override
    public void close() {
        run = false;
    }
}